#include"scheduler.h"

#include"vpm.h"
#include"mem.h"

#include<kernel/arch/x86/consts.h>

#include"canal.h"

#define PROCESS_ZONE_SIZE 4096
#define HANDLE_ZONE_SIZE 4096

Process *scheduler_ProcessZone = 0;
Process *scheduler_ProcessQueues[TOTAL_PRIORITIES];

static int scheduler_CurrentPriority;
volatile Process *scheduler_CurrentProcess = 0;

uint32_t scheduler_Tick;

void klogs(const char *str);
void klogc(int c);

static void klogp(void *p) {
	static char hex[16]="0123456789ABCDEF";
	
	uintptr_t i = (uintptr_t) p;
	klogc(hex[(i >> 28) & 0xF]);
	klogc(hex[(i >> 24) & 0xF]);
	klogc(hex[(i >> 20) & 0xF]);
	klogc(hex[(i >> 16) & 0xF]);
	klogc(hex[(i >> 12) & 0xF]);
	klogc(hex[(i >> 8) & 0xF]);
	klogc(hex[(i >> 4) & 0xF]);
	klogc(hex[(i >> 0) & 0xF]);
}

void scheduler_init() {
	uint32_t cr3;
	asm volatile("movl %%cr3, %0" : "=r" (cr3) :);
	
	scheduler_ProcessZone = vpm_map(cr3, VPM_KERNEL, -1, -1, PROCESS_ZONE_SIZE);
	kmemw1(scheduler_ProcessZone, 0, PROCESS_ZONE_SIZE);
}

static int consider_switch(Process *p) {
	if(p->waitType == WAIT_TYPE_SIGNAL) {
		uint32_t meta = 0;
		uint32_t canalId = -1;
		uint32_t which = is_signal_awaiting_to_concerning_link(p->id, p->waitValue, &meta, &canalId);
		if(which != 0xFFFFFFFF) {
			p->waitType = WAIT_TYPE_NONE;
			p->ctx.eax = which;
			p->ctx.edx = meta;
			p->ctx.ebx = canalId;
		} else return 0;
	} else if(p->waitType == WAIT_TYPE_TIME) {
		if(p->waitValue <= scheduler_Tick) {
			p->waitType = WAIT_TYPE_NONE;
		} else return 0;
	} else if(p->waitType == WAIT_TYPE_IRQ) {
		if(p->irqWaiting) {
			p->waitType = WAIT_TYPE_NONE;
			p->ctx.eip = p->irqWaiting;
			p->irqWaiting = 0;
		} else return 0;
	} else if(p->waitType == WAIT_TYPE_MINOR) {
		//asm("xchg %bx, %bx");
		return 0;
	}
	
	/*static char hex[16]="0123456789ABCDEF";
	klogc('0' + p->id);
	klogc(' ');
	klogc(hex[p->priority / PRIORITY_UNIT]);
	klogc('\n');*/
	
	scheduler_CurrentPriority = p->minipriority / PRIORITY_UNIT;
	scheduler_CurrentProcess = p;
	
	return 1;
}

void scheduler_switch() {
	if(scheduler_CurrentProcess == 0) {
		for(int p = 0; p < TOTAL_PRIORITIES; p++) {
			Process *q = scheduler_ProcessQueues[p];
			while(q) {
				if(consider_switch(q)) {
					return;
				}
				q = q->next;
			}
		}
		
		scheduler_CurrentProcess = 0;
		
		return;
	}
	
	for(int p = 0; p < scheduler_CurrentPriority; p++) {
		Process *q = scheduler_ProcessQueues[p];
		
		while(q) {
			if(consider_switch(q)) {
				return;
			}
			q = q->next;
		}
	}
	
	Process *p = scheduler_CurrentProcess;
	do {
		p = p->next;
		
		if(!p) {
			p = scheduler_ProcessQueues[scheduler_CurrentPriority];
		}
		
		if(consider_switch(p)) {
			return;
		}
	} while(p != scheduler_CurrentProcess);
	
	for(int p = scheduler_CurrentPriority + 1; p < TOTAL_PRIORITIES; p++) {
		Process *q = scheduler_ProcessQueues[p];
		
		while(q) {
			if(consider_switch(q)) {
				return;
			}
			q = q->next;
		}
	}
	
	scheduler_CurrentProcess = 0;
}

Process *scheduler_find(uint16_t id) {
	for(int p = 0; p < TOTAL_PRIORITIES; p++) {
		Process *q = scheduler_ProcessQueues[p];
		
		while(q) {
			if(q->id == id) {
				return q;
			}
			q = q->next;
		}
	}
	
	return NULL;
}

static uint16_t next_free_id() {
	for(uint32_t id = 1; id < 65536; id++) {
		if(!scheduler_find(id)) {
			return id;
		}
	}
	return 0;
}

Process *scheduler_spawn(int minipriority) {
	int found = 0;
	Process *p = scheduler_ProcessZone;
	
	for(size_t i = 0; i < PROCESS_ZONE_SIZE / sizeof(Process); i++) {
		if(p->id == 0) {
			found = 1;
			break;
		}
		
		p++;
	}
	
	if(!found) {
		return NULL;
	}
	
	kmemw1(p, 0, sizeof(*p));
	p->id = next_free_id();
	p->ctx.eflags = EFLAGS_RESET | EFLAGS_IF | EFLAGS_IOPL_3;
	p->ctx.cs = 24 | RPL_3;
	p->waitType = WAIT_TYPE_NONE;
	p->irqWaiting = 0;
	p->minipriority = minipriority;
	p->next = scheduler_ProcessQueues[minipriority / PRIORITY_UNIT];
	p->eventsRequest = 0;
	p->eventsSignal = 0;
	p->eventsClosed = 0;
	p->ownedCanal = NULL;
	p->guestLinks = NULL;
	
	scheduler_ProcessQueues[minipriority / PRIORITY_UNIT] = p;
	
	return p;
}

size_t scheduler_enslave(uint16_t ownerId, Process *child) {
	child->waitType = WAIT_TYPE_MINOR;
	child->ownerId = ownerId;
	
	// Should not be how it works.
	child->ctx.cr3 = vpm_create_space();
	
	return child->id;
}

int scheduler_unslave(size_t handle) {
	struct Process *proc = scheduler_find(handle);
	if(proc && proc->waitType == WAIT_TYPE_MINOR) {
		proc->waitType = WAIT_TYPE_NONE;
		return 0;
	}
	return -1;
}

#define SYS_STATE_X86_EIP 0
int process_abuse_state(struct Process *proc, size_t type, uintmax_t val) {
	switch(type) {
	case SYS_STATE_X86_EIP:
		proc->ctx.eip = val;
		break;
	default:
		return -1;
	}
	return 0;
}

#define KEV_TYPE_NONE 0
#define KEV_TYPE_LINK_REQUEST 1
#define KEV_TYPE_LINK_SIGNAL 2
#define KEV_TYPE_LINK_CLOSE 3
typedef struct __attribute__((packed)) {
	uint8_t type;
	uint32_t canalId;
	void *area;
	uint32_t meta;
} KEvent;

int process_search_event(Process *proc, KEvent *kev) {
	if(proc->eventsRequest) {
		kev->type = KEV_TYPE_LINK_REQUEST;
		
		Canal *c = proc->ownedCanal;
		while(c) {
			Link *l = c->links;
			
			while(l) {
				if(!(l->flags & LINK_FLAG_ACCEPTED)) {
					kev->canalId = l->canalId;
					
					return 1;
				}
				
				l = l->nextInCanal;
			}
			
			c = c->ownedNext;
		}
	}
	
	if(proc->eventsSignal) {
		kev->type = KEV_TYPE_LINK_SIGNAL;
		
		Link *l = proc->guestLinks;
		while(l) {
			if(l->flags & LINK_FLAG_H2G) {
				kev->area = signal_accept(proc, l, &kev->meta);
				
				return 1;
			}
			
			l = l->nextGuestLink;
		}
	}
	
	if(proc->eventsSignal) {
		kev->type = KEV_TYPE_LINK_SIGNAL;
		
		Canal *c = proc->ownedCanal;
		while(c) {
			Link *l = c->links;
			while(l) {
				if(l->flags & LINK_FLAG_G2H) {
					kev->area = signal_accept(proc, l, &kev->meta);
					
					return 1;
				}
				
				l = l->nextInCanal;
			}
			
			c = c->ownedNext;
		}
	}
	
	kev->type = KEV_TYPE_NONE;
	
	return 0;
}
